@@ -167,7 +167,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		return;
 	}
 
-	calls = at_util_parse_clcc(result);
+	calls = at_util_parse_clcc(result, NULL);
 
 	n = calls;
 	o = vd->calls;
@@ -253,8 +253,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		}
 	}
 
-	g_slist_foreach(vd->calls, (GFunc) g_free, NULL);
-	g_slist_free(vd->calls);
+	g_slist_free_full(vd->calls, g_free);
 
 	vd->calls = calls;
 
@@ -477,7 +476,7 @@ static void clcc_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	if (!ok)
 		return;
 
-	vd->calls = at_util_parse_clcc(result);
+	vd->calls = at_util_parse_clcc(result, NULL);
 
 	for (l = vd->calls; l; l = l->next)
 		ofono_voicecall_notify(vc, l->data);
@@ -1147,8 +1146,7 @@ static void at_voicecall_remove(struct ofono_voicecall *vc)
 	if (vd->vts_source)
 		g_source_remove(vd->vts_source);
 
-	g_slist_foreach(vd->calls, (GFunc) g_free, NULL);
-	g_slist_free(vd->calls);
+	g_slist_free_full(vd->calls, g_free);
 
 	ofono_voicecall_set_data(vc, NULL);
 
