@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2011-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -30,11 +30,36 @@
 
 static int qmimodem_init(void)
 {
+	qmi_devinfo_init();
+	qmi_netreg_init();
+	qmi_voicecall_init();
+	qmi_sim_legacy_init();
+	qmi_sim_init();
+	qmi_sms_init();
+	qmi_ussd_init();
+	qmi_gprs_init();
+	qmi_gprs_context_init();
+	qmi_radio_settings_init();
+	qmi_location_reporting_init();
+	qmi_netmon_init();
+
 	return 0;
 }
 
 static void qmimodem_exit(void)
 {
+	qmi_netmon_exit();
+	qmi_location_reporting_exit();
+	qmi_radio_settings_exit();
+	qmi_gprs_context_exit();
+	qmi_gprs_exit();
+	qmi_ussd_exit();
+	qmi_sms_exit();
+	qmi_sim_exit();
+	qmi_sim_legacy_exit();
+	qmi_voicecall_exit();
+	qmi_netreg_exit();
+	qmi_devinfo_exit();
 }
 
 OFONO_PLUGIN_DEFINE(qmimodem, "Qualcomm QMI modem driver", VERSION,
