@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -53,6 +53,10 @@ struct opl_operator {
 	guint8 id;
 };
 
+#define MF	1
+#define DF	2
+#define EF	4
+
 #define BINARY 0
 #define RECORD 1
 #define CYCLIC 3
@@ -66,56 +70,46 @@ struct opl_operator {
 #define ROOTMF 0x3F00
 
 static struct sim_ef_info ef_db[] = {
-{	0x2F05, ROOTMF, BINARY, 0,	ALW,	PIN	},
-{	0x2F06, ROOTMF, RECORD, 0,	ALW,	PIN	},
-{	0x2FE2, ROOTMF, BINARY, 10,	ALW,	NEV 	},
-{	0x4F20, 0x5F50, BINARY, 0,	PIN,	ADM	},
-{	0x6F05, 0x7F20, BINARY, 0,	ALW,	PIN	},
-{	0x6F06, 0x0000, RECORD, 0,	ALW,	ADM	},
-{	0x6F2C, 0x7F20, BINARY, 16,	PIN,	PIN	},
-{	0x6F30, 0x7F20, BINARY, 0,	PIN,	PIN	},
-{	0x6F32, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F37, 0x7F20, BINARY, 3,	PIN,	PIN2	},
-{	0x6F38, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F39, 0x7F20, CYCLIC, 3,	PIN,	PIN2	},
-{	0x6F3B, 0x7F10, RECORD, 0,	PIN,	PIN2	},
-{	0x6F3E, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F3F, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F40, 0x7F10, RECORD, 0,	PIN,	PIN	},
-{	0x6F41, 0x7F20, BINARY, 5,	PIN,	PIN2	},
-{	0x6F42, 0x7F10, RECORD, 0,	PIN,	PIN	},
-{	0x6F44, 0x7F10, CYCLIC, 0,	PIN,	PIN	},
-{	0x6F45, 0x7F20, BINARY, 0,	PIN,	PIN	},
-{	0x6F46, 0x7F20, BINARY, 17,	ALW,	ADM	},
-{	0x6F48, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F49, 0x7F10, RECORD, 0,	PIN,	ADM	},
-{	0x6F4D, 0x7F20, RECORD, 0,	PIN,	PIN2	},
-{	0x6F50, 0x7F20, BINARY, 0,	PIN,	PIN	},
-{	0x6F51, 0x7F20, RECORD, 0,	PIN,	ADM	},
-{	0x6F53, 0x7F20, BINARY, 14,	PIN,	PIN	},
-{	0x6F56, 0x0000, BINARY, 0,	PIN,	PIN2	},
-{	0x6F60, 0x7F20, BINARY, 0,	PIN,	PIN	},
-{	0x6F61, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F62, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6F73, 0x0000, BINARY, 14,	PIN,	PIN	},
-{	0x6F7B, 0x7F20, BINARY, 0,	PIN,	PIN	},
-{	0x6F7E, 0x7F20, BINARY, 11,	PIN,	PIN	},
-{	0x6FAD, 0x7F20, BINARY, 0,	ALW,	ADM	},
-{	0x6FAE, 0x7F20, BINARY, 1,	ALW,	ADM	},
-{	0x6FB7, 0x7F20, BINARY, 0,	ALW,	ADM	},
-{	0x6FC5, 0x7F20, RECORD, 0,	ALW,	ADM	},
-{	0x6FC6, 0x7F20, RECORD, 0,	ALW,	ADM	},
-{	0x6FC7, 0x7F20, RECORD, 0,	PIN,	PIN	},
-{	0x6FC9, 0x7F20, RECORD, 0,	PIN,	PIN	},
-{	0x6FCA, 0x7F20, RECORD, 0,	PIN,	PIN	},
-{	0x6FCB, 0x7F20, RECORD, 16,	PIN,	PIN	},
-{	0x6FCD, 0x7F20, BINARY, 0,	PIN,	ADM	},
-{	0x6FD9, 0x0000, BINARY, 0,	PIN,	ADM	},
-{	0x6FDB, 0x0000, BINARY, 1,	PIN,	ADM	},
-{	0x6FDC, 0x0000, BINARY, 1,	PIN,	ADM	},
-{	0x6FDE, 0x0000, BINARY, 0,	ALW,	ADM	},
-{	0x6FDF, 0x0000, RECORD, 0,	ALW,	ADM	},
-{	0x6FE3, 0x0000, BINARY, 18,	PIN,	PIN	},
+{	0x2F05, ROOTMF, ROOTMF, EF, BINARY, 0,		ALW,	PIN	},
+{	0x2FE2, ROOTMF, ROOTMF, EF, BINARY, 10,		ALW,	NEV	},
+{	0x4F20, 0x5F50, 0x5F50, EF, BINARY, 0,		PIN,	ADM	},
+{	0x5F50, 0x7F10, 0x7F10, DF, 0, 0,		PIN,	ADM	},
+{	0x6F05, 0x7F20, 0x7FFF, EF, BINARY, 0,		ALW,	PIN	},
+{	0x6F07, 0x7F20, 0x7FFF, EF, BINARY, 9,		PIN,	ADM	},
+{	0x6F11, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	PIN	},
+{	0x6F13, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	PIN	},
+{	0x6F14, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	ADM	},
+{	0x6F15, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	PIN	},
+{	0x6F16, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	ADM	},
+{	0x6F17, 0x7F20, 0x7FFF, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6F18, 0x7F20, 0x7FFF, EF, BINARY, 10,		PIN,	ADM	},
+{	0x6F19, 0x7F20, 0x7FFF, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6F38, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	ADM	},
+{	0x6F3A, 0x7F10, 0x7F10, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6F3B, 0x7F10, 0x7FFF, EF, RECORD, 0,		PIN,	PIN2	},
+{	0x6F40, 0x7F10, 0x7FFF, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6F45, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	PIN	},
+{	0x6F46, 0x7F20, 0x7FFF, EF, BINARY, 17,		ALW,	ADM	},
+{	0x6F48, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	ADM	},
+{	0x6F49, 0x7F10, 0x7FFF, EF, RECORD, 0,		PIN,	ADM	},
+{	0x6F4D, 0x7F20, 0x7FFF, EF, RECORD, 0,		PIN,	PIN2	},
+{	0x6F50, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	PIN	},
+{	0x6F56, 0x0000, 0x7FFF, EF, BINARY, 0,		PIN,	PIN2	},
+{	0x6F57,	0x7F20,	0x7FFF,	EF, BINARY, 0,		PIN,	PIN2	},
+{	0x6FAD, 0x7F20, 0x7FFF, EF, BINARY, 0,		ALW,	ADM	},
+{	0x6FAE, 0x7F20, 0x0000, EF, BINARY, 1,		ALW,	ADM	},
+{	0x6FB7, 0x7F20, 0x7FFF, EF, BINARY, 0,		ALW,	ADM	},
+{	0x6FC5, 0x7F20, 0x7FFF, EF, RECORD, 0,		ALW,	ADM	},
+{	0x6FC6, 0x7F20, 0x7FFF, EF, RECORD, 0,		ALW,	ADM	},
+{	0x6FC7, 0x7F20, 0x7FFF, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6FC9, 0x7F20, 0x7FFF, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6FCA, 0x7F20, 0x7FFF, EF, RECORD, 0,		PIN,	PIN	},
+{	0x6FCB, 0x7F20, 0x7FFF, EF, RECORD, 16,		PIN,	PIN	},
+{	0x6FCD, 0x7F20, 0x7FFF, EF, BINARY, 0,		PIN,	ADM	},
+{	0x6FDE, 0x7F20, 0x7FFF, EF, BINARY, 0,		ALW,	ADM	},
+{	0x7F10, ROOTMF, ROOTMF, DF, 0, 0,		0,	0	},
+{	0x7F20, ROOTMF, ROOTMF, DF, 0, 0,		0,	0	},
+{	0x7FFF, 0x0000, ROOTMF, DF, 0, 0,		0,	0	}
 };
 
 void simple_tlv_iter_init(struct simple_tlv_iter *iter,
@@ -232,7 +226,7 @@ gboolean comprehension_tlv_iter_next(struct comprehension_tlv_iter *iter)
 
 	/*
 	 * ETSI TS 101.220, Section 7.1.1.2
-	 * 
+	 *
 	 * If byte 1 of the tag is equal to 0x7F, then the tag is encoded
 	 * on the following two bytes, with bit 8 of the 2nd byte of the tag
 	 * being the CR flag.
@@ -723,11 +717,10 @@ gboolean comprehension_tlv_builder_set_length(
 	unsigned char *tlv = builder->pdu + builder->pos;
 	unsigned int tag_size = CTLV_TAG_FIELD_SIZE(tlv[0]);
 	unsigned int len_size, new_len_size;
-	unsigned int ctlv_len, new_ctlv_len;
+	unsigned int new_ctlv_len;
 	unsigned int len;
 
 	len_size = CTLV_LEN_FIELD_SIZE(tlv[tag_size]);
-	ctlv_len = tag_size + len_size + builder->len;
 	new_len_size = BTLV_LEN_FIELD_SIZE_NEEDED(new_len);
 	new_ctlv_len = tag_size + new_len_size + new_len;
 
@@ -775,12 +768,14 @@ static char *sim_network_name_parse(const unsigned char *buffer, int length,
 					gboolean *add_ci)
 {
 	char *ret = NULL;
-	unsigned char *endp;
 	unsigned char dcs;
 	int i;
 	gboolean ci = FALSE;
+	unsigned char *unpacked_buf;
+	long num_char, written;
+	int spare_bits;
 
-	if (length < 1)
+	if (length < 2)
 		return NULL;
 
 	dcs = *buffer++;
@@ -795,11 +790,18 @@ static char *sim_network_name_parse(const unsigned char *buffer, int length,
 
 	switch (dcs & (7 << 4)) {
 	case 0x00:
-		endp = memchr(buffer, 0xff, length);
-		if (endp)
-			length = endp - buffer;
-		ret = convert_gsm_to_utf8(buffer, length,
-				NULL, NULL, 0xff);
+		spare_bits = dcs & 0x07;
+		num_char = (length * 8 - spare_bits) / 7;
+
+		unpacked_buf = unpack_7bit(buffer, length, 0, FALSE,
+						num_char, &written, 0);
+		if (unpacked_buf == NULL)
+			break;
+
+		ret = convert_gsm_to_utf8(unpacked_buf, written, NULL, NULL, 0);
+
+		g_free(unpacked_buf);
+
 		break;
 	case 0x10:
 		if ((length % 2) == 1) {
@@ -975,13 +977,18 @@ gboolean sim_spdi_lookup(struct sim_spdi *spdi,
 
 void sim_spdi_free(struct sim_spdi *spdi)
 {
-	g_slist_foreach(spdi->operators, (GFunc)g_free, NULL);
-	g_slist_free(spdi->operators);
+	if (spdi == NULL)
+		return;
+
+	g_slist_free_full(spdi->operators, g_free);
 	g_free(spdi);
 }
 
 static void pnn_operator_free(struct sim_eons_operator_info *oper)
 {
+	if (oper == NULL)
+		return;
+
 	g_free(oper->info);
 	g_free(oper->shortname);
 	g_free(oper->longname);
@@ -1072,13 +1079,15 @@ void sim_eons_free(struct sim_eons *eons)
 {
 	int i;
 
+	if (eons == NULL)
+		return;
+
 	for (i = 0; i < eons->pnn_max; i++)
 		pnn_operator_free(eons->pnn_list + i);
 
 	g_free(eons->pnn_list);
 
-	g_slist_foreach(eons->opl_list, (GFunc)g_free, NULL);
-	g_slist_free(eons->opl_list);
+	g_slist_free_full(eons->opl_list, g_free);
 
 	g_free(eons);
 }
@@ -1288,6 +1297,72 @@ struct sim_ef_info *sim_ef_db_lookup(unsigned short id)
 	return result;
 }
 
+unsigned int sim_ef_db_get_path_2g(unsigned short id, unsigned char out_path[])
+{
+	struct sim_ef_info *info;
+	unsigned int nelem = sizeof(ef_db) / sizeof(struct sim_ef_info);
+	unsigned char path[6];
+	int i = 0;
+	int j;
+
+	info = bsearch(GUINT_TO_POINTER((unsigned int) id), ef_db, nelem,
+				sizeof(struct sim_ef_info), find_ef_by_id);
+	if (info == NULL)
+		return 0;
+
+	path[i++] = info->parent2g & 0xff;
+	path[i++] = info->parent2g >> 8;
+
+	while (info->parent2g != ROOTMF) {
+		info = bsearch(GUINT_TO_POINTER((unsigned int) info->parent2g),
+				ef_db, nelem, sizeof(struct sim_ef_info),
+				find_ef_by_id);
+		if (info == NULL)
+			return 0;
+
+		path[i++] = info->parent2g & 0xff;
+		path[i++] = info->parent2g >> 8;
+	}
+
+	for (j = 0; j < i; j++)
+		out_path[j] = path[i - j - 1];
+
+	return i;
+}
+
+unsigned int sim_ef_db_get_path_3g(unsigned short id, unsigned char out_path[])
+{
+	struct sim_ef_info *info;
+	unsigned int nelem = sizeof(ef_db) / sizeof(struct sim_ef_info);
+	unsigned char path[6];
+	int i = 0;
+	int j;
+
+	info = bsearch(GUINT_TO_POINTER((unsigned int) id), ef_db, nelem,
+				sizeof(struct sim_ef_info), find_ef_by_id);
+	if (info == NULL)
+		return 0;
+
+	path[i++] = info->parent3g & 0xff;
+	path[i++] = info->parent3g >> 8;
+
+	while (info->parent3g != ROOTMF) {
+		info = bsearch(GUINT_TO_POINTER((unsigned int) info->parent3g),
+				ef_db, nelem, sizeof(struct sim_ef_info),
+				find_ef_by_id);
+		if (info == NULL)
+			return 0;
+
+		path[i++] = info->parent3g & 0xff;
+		path[i++] = info->parent3g >> 8;
+	}
+
+	for (j = 0; j < i; j++)
+		out_path[j] = path[i - j - 1];
+
+	return i;
+}
+
 gboolean sim_parse_3g_get_response(const unsigned char *data, int len,
 					int *file_len, int *record_len,
 					int *structure, unsigned char *access,
@@ -1466,6 +1541,14 @@ gboolean sim_sst_is_active(unsigned char *efsst, unsigned char len,
 	return (efsst[index / 4] >> (((index % 4) * 2) + 1)) & 1;
 }
 
+gboolean sim_cphs_is_active(unsigned char *cphs, enum sim_cphs_service index)
+{
+	if (index >= 2 * 4u)
+		return FALSE;
+
+	return ((cphs[index / 4] >> ((index % 4) * 2)) & 3) == 3;
+}
+
 GSList *sim_parse_app_template_entries(const unsigned char *buffer, int len)
 {
 	GSList *ret = NULL;
@@ -1487,6 +1570,8 @@ GSList *sim_parse_app_template_entries(const unsigned char *buffer, int len)
 
 		memcpy(app.aid, aid, app.aid_len);
 
+		app.type = (app.aid[5] << 8) & app.aid[6];
+
 		/* Find the label (optional) */
 		label = ber_tlv_find_by_tag(dataobj, 0x50, dataobj_len,
 						&label_len);
@@ -1524,3 +1609,132 @@ error:
 
 	return NULL;
 }
+
+static int build_authenticate(unsigned char *buffer, const unsigned char *rand,
+		const unsigned char *autn)
+{
+	int pos = 0;
+
+	buffer[pos++] = 0x00;
+	buffer[pos++] = 0x88;
+	buffer[pos++] = 0x00;
+	buffer[pos++] = autn ? 0x81 : 0x80;
+	buffer[pos++] = autn ? 0x22 : 0x11;
+	buffer[pos++] = 0x10;
+	memcpy(buffer + pos, rand, 16);
+	pos += 16;
+
+	if (autn) {
+		buffer[pos++] = 0x10;
+		memcpy(buffer + pos, autn, 16);
+		pos += 16;
+	}
+
+	buffer[pos++] = 0x00;
+
+	return pos;
+}
+
+int sim_build_umts_authenticate(unsigned char *buffer, int len,
+		const unsigned char *rand, const unsigned char *autn)
+{
+	if (len < 40 || !rand || !autn)
+		return FALSE;
+
+	return build_authenticate(buffer, rand, autn);
+}
+
+int sim_build_gsm_authenticate(unsigned char *buffer, int len,
+		const unsigned char *rand)
+{
+	if (len < 22 || !rand)
+		return FALSE;
+
+	return build_authenticate(buffer, rand, NULL);
+}
+
+gboolean sim_parse_umts_authenticate(const unsigned char *buffer,
+		int len, const unsigned char **res, const unsigned char **ck,
+		const unsigned char **ik, const unsigned char **auts,
+		const unsigned char **kc)
+{
+	if (len < 16 || !buffer)
+		return FALSE;
+
+	switch (buffer[0]) {
+	case 0xdb:
+		/* 'DB' + '08' + RES(16) + '10' + CK(32) + '10' + IK(32) = 43 */
+		if (len < 43)
+			goto umts_end;
+
+		/* success */
+		if (buffer[1] != 0x08)
+			goto umts_end;
+
+		*res = buffer + 2;
+
+		if (buffer[10] != 0x10)
+			goto umts_end;
+
+		*ck = buffer + 11;
+
+		if (buffer[27] != 0x10)
+			goto umts_end;
+
+		*ik = buffer + 28;
+
+		if (len >= 53 && kc) {
+			if (buffer[44] != 0x08)
+				goto umts_end;
+
+			*kc = buffer + 45;
+		} else {
+			*kc = NULL;
+		}
+
+		*auts = NULL;
+
+		break;
+	case 0xdc:
+		/* 'DC' + '0E' + AUTS(14) = 16 */
+		if (len < 16)
+			goto umts_end;
+
+		/* sync error */
+		if (buffer[1] != 0x0e)
+			goto umts_end;
+
+		*auts = buffer + 2;
+
+		break;
+	default:
+		goto umts_end;
+	}
+
+	return TRUE;
+
+umts_end:
+	return FALSE;
+}
+
+gboolean sim_parse_gsm_authenticate(const unsigned char *buffer, int len,
+		const unsigned char **sres, const unsigned char **kc)
+{
+	if (len < 14 || !buffer)
+		goto gsm_end;
+
+	if (buffer[0] != 0x04)
+		goto gsm_end;
+
+	*sres = buffer + 1;
+
+	if (buffer[5] != 0x08)
+		goto gsm_end;
+
+	*kc = buffer + 6;
+
+	return TRUE;
+
+gsm_end:
+	return FALSE;
+}
